/* Twilight clock - program code */
/* CC BY-NC-SA Jeroen Brinkman */
/* Version v7 dd 14-4-2022 */

#include <LiquidCrystal_I2C.h>
#include <Wire.h>
#include "RTClib.h"
#include "EEPROM.h"
#if defined(ARDUINO) && ARDUINO >= 100
#define printByte(args)  write(args);
#else
#define printByte(args)  print(args,BYTE);
#endif

// adjust the following #defines when needed 
#define REGION "EU"    // Use European standard for Daylight saving time calculation
//#define REGION "XX"    // Change REGION to XX if not living in the EU
#define ROTA 7 // Use ROTA 6 if right turn does not increase time
#define ROTB 6 // Use ROTB 7 if right turn does not increase time

#define T_FROM  "Van-"        // Dutch text, exactly 4 characters, change if needed
#define T_UNTIL "Tot-"        // Dutch text, exactly 4 characters, change if needed
#define T_OFF  "Uit"          // Dutch text, exactly 3 characters, change if needed
#define T_ON   "Aan"          // Dutch text, exactly 3 characters, change if needed
#define T_SET  "Zet   "       // Dutch text, exactly 6 characters, change if needed
#define T_COR  " Correct"     // Dutch text, exactly 8 characters, change if needed
#define T_SAM  "Zelfdetijd  " // Dutch text, exactly 12 characters, change if needed
#define T_WIN  "Wintertijd  " // Dutch text, exactly 12 characters, change if needed
#define T_SUM  "Zomertijd   " // Dutch text, exactly 12 characters, change if needed
#define T_INT  "Tussentijd  " // Dutch text, exactly 12 characters, change if needed
#define T_NAME " Schemerklok v7 " // Dutch text, exactly 16 characters, change if needed

/*
#define T_FROM  " On-"        // English text, exactly 4 characters, change if needed
#define T_UNTIL "Off-"        // English text, exactly 4 characters, change if needed
#define T_OFF  "Off"          // English text, exactly 3 characters, change if needed
#define T_ON   "On "          // English text, exactly 3 characters, change if needed
#define T_SET  "Set   "       // English text, exactly 6 characters, change if needed
#define T_COR  " Correct"     // English text, exactly 8 characters, change if needed
#define T_SAM  "No DST      " // English text, exactly 12 characters, change if needed
#define T_WIN  "Wintertime  " // English text, exactly 12 characters, change if needed
#define T_SUM  "Summertime  " // English text, exactly 12 characters, change if needed
#define T_INT  "Intertime   " // English text, exactly 12 characters, change if needed
#define T_NAME "Twilightclock v7" // English text, exactly 16 characters, change if needed
end of adjust
*/

#define AWAKE 120000   // Number of milliseconds before go to sleep (twilight mode)
#define WAIT   60000   // Number of milliseconds that the twilight must be detected continuously
#define MINUTE 60000   // Milliseconds that represents a minute
#define PRESS 2000     // Milliseconds to press to detect a long press
#define T_IS "-"       // sensor result sign
#define T_DAT "/"      // date seperation sig4
#define T_DP ":"       // time seperation sign
#define T_SEP "|"      // screen separator


#define OVER    1      // place in work[] where the manual override is stored
#define HASH    2      // place in work[] where the hash code is stored
#define FHR     3      // place in work[] where the from hours are stored
#define FMIN    4      // place in work[] where the from minutes are stored
#define UHR     5      // place in work[] where the until hours are stored
#define UMIN    6      // place in work[] where the until minutes are stored
#define MODE    7      // place in work[] where the mode is stored
#define HR      8      // place in work[] where the rtc actual houres are stored
#define MIN     9      // place in work[] where the rtc actual minutes are stored
#define DAY    10      // place in work[] where the rtc actual day is stored
#define MONTH  11      // place in work[] where the rtc actual month is stored
#define YEAR   12      // place in work[] where the rtc actual year is stored
#define STATUS 13      // place in work[] where the Daylight status is stored
#define DRIFT  14      // place in work[] where the drift correction is stored
#define FROM   15      // place in work[] where the from time expressed in minutes after midnight is stored
#define UNTIL  16      // place in work[] where the until time expressed in minutes after midnight is stored
#define COL     0      // Position in pos[] where the display column is stored
#define ROW     1      // Position in pos[] where the display row is stored
#define MAX     2      // Position in pos[] where the Maximum is stored
#define STOR    3      // Position in pos[] where the EEPROM position is stored

#define SLEEP   1      // Status for sleep/dimmed mode
#define RUN     2      // Status for running mode
#define NONE    0      // no override
#define NORMAL  0      // normal operation: sensor & clock
#define OFF     1      // lights manual set off or system set off
#define ON      2      // lights manual set on
#define CLOCK   2      // clock only
#define SENSOR  3      // sensor only
#define WINTER  1      // winter time
#define SUMMER  2      // Summer time
#define INTER   3      // Intermediate time
#define ZERO   99      // for drift adjustment in use to avoid negative numbers
#define DRIDAY 14      // Minimal number of days between drift adjustments

#define IN_RUNNING_MODE (stat < 3)
#define IN_EDITING_MODE (stat > 2)

#define SIZE    17     // Size of work[] and pos[]
#define ESIZE    8     // Size of EEPROM 
#define BOUNCE  16     // Number of debouncing samples
#define MARGIN  30     // Margin for backup twilight calculation
#define PUSH     5     // Input push button
#define ROTP     8     // Input rotary button
#define TWILIGHT 9     // input twilight sensor
#define LIGHTON 10     // Output light on or off
#define HEART   13     // On-board heartbeat

RTC_DS1307 RTC;
LiquidCrystal_I2C lcd(0x27, 16, 2); // set the LCD address to 0x27 for a 16 chars and 2 line display

byte chr_bell[8]  = {0x00,0x04,0x0E,0x0E,0x0E,0x1F,0x00,0x04}; // ) ( 0. switch clock and twilight detection 
byte chr_nope[8]  = {0x00,0x00,0x12,0x0C,0x0C,0x12,0x00,0x00}; //  X  1. sensor switched off 
byte chr_clock[8] = {0x00,0x0E,0x15,0x17,0x11,0x0E,0x00,0x00}; //  O  2. only switch clock   
byte chr_sun[8]   = {0x00,0x15,0x0E,0x1B,0x0E,0x15,0x00,0x00}; //  *  3. only light detection 
byte chr_on[8]    = {0x00,0x00,0x1E,0x1E,0x1E,0x1E,0x00,0x00}; // [o] 4. sensor on 
byte chr_off[8]   = {0x00,0x00,0x1E,0x12,0x12,0x1E,0x00,0x00}; // [ ] 5. sensor off or system off 
byte chr_glass[8] = {0x1F,0x0A,0x0E,0x04,0x0E,0x0E,0x1F,0x00}; // ><  6. sensor defect, backup is used 
byte chr_half[8]  = {0x00,0x00,0x00,0x0C,0x0C,0x00,0x00,0x00}; //  o  7. light less than 1 minute detected


const int pos[SIZE][4] = {{6,0,0,0},{6,1,0,0},{3,1,255,1},{11,0,23,2},{14,0,59,3},{11,1,23,4},
{14,1,59,5},{5,0,3,6},{0,0,23,0},{3,0,59,0},{0,0,31,0},{3,0,12,0},{6,0,99,0},{0,1,2,7},{12,1,2*ZERO,8},{0,0,0,0},{0,0,0,0}};

const int mon[13] = {0,31,28,31,30,31,30,31,31,30,31,30,31}; // days in a month
int work[SIZE]; 
int action, stat, pt, rot, hash, signalp, signalr, signalA, signalB, detect, sum, i, j, failhour; // integers
unsigned long nu, check1, check2; // daylight saving time and twilight sensor backup
unsigned long pressed, gotosleep, dark, mincheck; // milliseconds
bool new_signal, clck, twilight, signalt, adjusted, lastlight, newlight, light, setdate, rewrite, sensorfail, clockfail, done; // booleans
DateTime now, oldnow, temp, lastadjust;


/* Description used non-local variables and arrays
Pos:
  For each of the statuses the position on the screen and the max allowed value
    Col Row Max
    [0] [1] [2]
  0   6   0   0  - position vertical line first line
  1   6   1   0  - position vertical line second line
  2   3   1 255  - position light status and max value hash
  3  11   0  23  - start hours
  4  14   0  59  - start minutes
  5  11   1  23  - until hours
  6  14   1  59  - until minutes
  7   5   0   3  - running mode
  8   0   0  23  - actual time hours
  9   3   0  59  - actual time minutes
 10   0   0  31  - day
 11   3   0  12  - month
 12   6   0  99  - year
 13   0   1   2  - status daylight saving time
 
EEPROM:
  0, default value 123 means that the EEPROM has been written once
  1 pt, default is 1
  pt+1 Hashtotal
  pt+2 Start hours
  pt+3 Start minutes
  pt+4 End hours
  pt+5 End minutes
  pt+6 Running mode
  pt+7 status daylight saving time 

Work: (* is stored in EEPROM)
  0  None
  1  Manual override
  2* Hashtotal
  3* Start hours
  4* Start minutes
  5* End hours
  6* End minutes
  7* Running mode
  8  Time hours
  9  Time minutes
 10  Date day
 11  Date month
 12  Date year
 13* Status daylight saving time
 14* Drift correction in seconds per day + 30
 15  From total minutes
 16  Until total minutes

  Manual override (work[OVER])
  0 running conform Mode
  1 lights manual set off
  2 lights manual set on
  
  Running mode (work[MODE])
  0 normal operation: sensor & clock
  1 system off
  2 clock only
  3 sensor only
  
  Status daylight saving time: (work[STATUS])
  0 daylight saving time off
  1 clock shows winter time
  2 clock shows summer time
  3 summer time has been changed to winter time
   
Action:
  0 None
  1 Toggle settings/run
  2 Manual lights on/off
  3 Value --
  4 Value ++
  5 Select
  6 Adjust date (future use)
  7 Show help text (future use)

State (stat):
  0 Error detected
  1 Running & sleeping
  2 Running & gotosleep
  3 Adjust start hours
  4 Adjust start minutes
  5 Adjust end hours
  6 Adjust end minutes
  7 Adjust run/off
  8 Adjust time hours
  9 Adjust time minutes
 10 Adjust date day
 11 Adjust date month
 12 Adjust date year
 13 Adjust status daylight saving time

Global variables:
  pt: Pointer to the first location in EEPROM where data is stored. If EEPROM becomes faulty, new (higher) memory positions are set
  checked: is it time to check the clock again? 
  gotosleep: moment the display must go to sleep
  twilight: has the twilight come?
  clck: is the current time between the from and until time
  new_signal: has de rotary knob been back to the null position?
  twilight: has de twilight sensor detected twilight?
  dark: time when if twilight is continuous dtected it is considered dark
  adjusted: has the clock time been adjusted in the settings?
  lastlight: what was the last status of the twilight clock?
  newlight: what should the light be based on the twilight clock and manual override
  light: are the lights currently on or off?
  setdate: if true, the date is edited instead of the settings
  rewrite: relocate the permanent stored data in the EEPROM, because there are read detects
  failhour: the hour the sensor must detect a change.
  sensorfail: true when the twilight sensor is not avalable or not working
  clockfail: true when the clock is not running and the Arduino takes over
  mincheck: when is the new check if the RTC is running
  oldnow: time aproximately a minute ago
  lastadjust: the last date and time the clock has been adjusted
  done: the time has been corrected for drift on this day
*/

void setup() {
  
/* ------- Setup -------  */
  
  Serial.begin(9600);
  Wire.begin();
  Serial.println("------------ New start ------------");

  /* define I/O  */
  pinMode(PUSH, INPUT_PULLUP);
  pinMode(ROTA, INPUT_PULLUP);
  pinMode(ROTB, INPUT_PULLUP);
  pinMode(ROTP, INPUT_PULLUP);
  pinMode(TWILIGHT, INPUT);
  pinMode(LIGHTON, OUTPUT);
  pinMode(HEART,  OUTPUT); // blinking led showing the programma's hartbeat

  /* initialize variables  */
  stat = 2; action = 0; twilight = 0; detect = 0; 
  light = false; lastlight = false; twilight = false; clck = false; adjusted = false; done = false;
  new_signal = true; setdate = false; sensorfail = false; clockfail = false;
  signalt = digitalRead(TWILIGHT);
  gotosleep = millis() + AWAKE;
  mincheck = millis() + MINUTE;
  work[OVER] = NORMAL;

  /* initialize the lcd  */
  lcd.init();
  lcd.backlight();
  lcd.createChar(0, chr_bell);
  lcd.createChar(1, chr_nope); 
  lcd.createChar(2, chr_clock);
  lcd.createChar(3, chr_sun);
  lcd.createChar(4, chr_on);
  lcd.createChar(5, chr_off);
  lcd.createChar(6, chr_glass); 
  lcd.createChar(7, chr_half);
  lcd.setCursor(0, 0); lcd.print(T_NAME);
  lcd.setCursor(0, 1); lcd.print("Jeroen Brinkman"); 
  delay(3000);
  
  /* initialize the rtc  */
  RTC.begin();
  if (!RTC.isrunning()) {
    Serial.println("RTC is NOT running!");
    RTC.adjust(DateTime(__DATE__, __TIME__));
  } 
  now = RTC.now();
   
  /* Initialize EEPROM if neccessary  */
  Serial.print("EEPROM[0]="); Serial.println(EEPROM.read(0));
  if (EEPROM.read(0) != 123) { // if first time use of EEPROM then
    pt = 1;
    for (i = 0; i < SIZE; i++) work[i] = 0;
    work[FHR] = 7; work[FMIN] = 28;
    work[UHR] = 22; work[UMIN] = 38;
    work[MODE] = NORMAL; work[DRIFT] = ZERO;
    work[STATUS] = NONE; 
    if ((REGION == "EU")) { // determine the daylight saving status
      temp = DateTime(now.year(), 4, 1, 1, 1, 1); // 1 april yyyy 1:01:01
      check1 = 33102L - ((temp.dayOfTheWeek() - 1) * 100); // last sunday of march march in format mmddhh
      temp = DateTime(now.year(), 11, 1, 1, 1, 1); // 1 november yyyy 1:01:01
      check2 = 103103L - ((temp.dayOfTheWeek() - 1) * 100); // last sunday of october in format mmddhh
      nu = (now.month()*100); nu = nu * 100; nu = nu + (now.day()*100) + now.hour(); // other construction gives wrong value during run-time
      work[STATUS] = SUMMER; if ((nu < check1) || (nu >= check2)) work[STATUS] = WINTER; 
    }
    work[HASH] = work[FHR] + work[FMIN] + work[UHR] + work[UMIN] + work[MODE] + work[STATUS] + work[DRIFT];
    for (i = 0; i < SIZE; i++) { // fill the EEPROM with work[]
      Serial.print("work["); Serial.print(i); Serial.print("]="); Serial.println(work[i]);
      if (pos[i][STOR] != 0) { // value need to be stored in EEPROM
        EEPROM.update(pt+pos[i][STOR], work[i]); 
      }
    }
    EEPROM.update(0, 123); // EEPROM has been initialized
    EEPROM.update(1, 1); // location of data in EEPROM
    Serial.println("EEPROM initialized"); 
  }
  pt = EEPROM.read(1);   

  /* Retrieve the EEPROM information  */  
  Serial.print("Pointer value:"); Serial.println(pt);
  sum = 0; rewrite = false;
  for (i = 0; i < SIZE; i++) { // fill work{} from the EEPROM
    if (pos[i][STOR] == 0) { // value is not stored in EEPROM
      work[i] = 0;
    } else {
      work[i] = EEPROM.read(pt+pos[i][STOR]); // read the EEPROM values into work[]
      if ((work[i] > pos[i][MAX]) || (work[i] < 0)) {
        rewrite = true; // if the stored value is not within the range, it must be wrong
        work[i] = 0; // set the value to zero
      }
      sum += work[i]; // sum the stored values
      Serial.print("work["); Serial.print(i); Serial.print("]="); Serial.print(work[i]); Serial.print("=EEPROM["); Serial.print(pt+pos[i][STOR]); Serial.println("] ");
    }  
  }
   sum = int(sum/2); // because the sum is also added
  Serial.print("Calculated checksum:"); Serial.println(sum);
  if (work[HASH] != sum) rewrite = true; // if the stored values do not equal the checksum then there is a problem
  if (sum == 0) rewrite = true;          // if the checksum = 0 then there is a problem
  Serial.print("Stored checksum:"); Serial.println(work[HASH]); 
  
  /* Fill work[] with time info  */    
  work[HR] = now.hour(); work[MIN] = now.minute();
  work[DAY] = now.day(); work[MONTH] = now.month(); work[YEAR] = now.year() - 2000;
  Serial.print("RTC actual date:"); Serial.print(work[DAY]); Serial.print(T_DAT); Serial.print(work[MONTH]); Serial.print(T_DAT); Serial.println(work[YEAR]);
  Serial.print("RTC actual time:"); Serial.print(work[HR]); Serial.print(T_DP); Serial.println(work[MIN]);
  work[FROM] = (work[FHR] * 60) + work[FMIN];
  work[UNTIL] = (work[UHR] * 60) + work[UMIN];
  
  failhour = now.hour() - 2;    
  if (failhour < 0) failhour = failhour + 24; // set the hour when failure is eminent with no changing twilight sensor value
  oldnow = now;
  mincheck = millis() + MINUTE;
}

void loop() {
  digitalWrite(HEART, (millis() / 1000) % 2); //1s heartbeat for the onboard led
  
  /* check if the rtc is running and if not use Arduino as a backup clock */
  
  now = RTC.now();
  /* Adjust the clock daily for drift */
  if ((now.hour() == 1) && (now.minute() == 0) && (now.second() == 0) && (!done) && (now != oldnow)) { // at 1:00:00 if clock is running an not already done
    Serial.print("Drift correction, old time="); Serial.print(now.hour()); Serial.print(T_DP); Serial.print(now.minute()); Serial.print(T_DP); Serial.println(now.second()); 
    i = (work[DRIFT]-ZERO) / 60; 
    j = (work[DRIFT]-ZERO) % 60;
    if (((work[DRIFT]-ZERO) > 0) && ((work[DRIFT]-ZERO) <= (2*ZERO))) {
      RTC.adjust(DateTime(now.year(), now.month(), now.day(), now.hour(), i, j)); // correct the time positive for drift
      Serial.println("Positive drift correction");
    }
    if (((work[DRIFT]-ZERO) < 0 ) && ((work[DRIFT]-ZERO) >= -(2*ZERO))) {
      i = 59 + i; 
      j = 60 + j;
      RTC.adjust(DateTime(now.year(), now.month(), now.day(), 0, i, j)); // correct the time negative for drift
      Serial.println("Negative drift correction");
    }
    now = RTC.now();
    Serial.print("work[DRIFT]="); Serial.print(work[DRIFT]); Serial.print(", "); Serial.print(-(2*ZERO)); Serial.print(", i="); Serial.print(i); Serial.print(", j="); Serial.println(j);     
    Serial.print(work[DRIFT]-ZERO); Serial.print("s, new time="); Serial.print(now.hour()); Serial.print(T_DP); Serial.print(now.minute()); Serial.print(T_DP); Serial.println(now.second()); 
    done = true;
  }
  if ((now.hour() == 2) && (now.minute() == 0) && (now.second() == 0)) done = false; // at 2:00:00 reset for the next day
        
  if ((mincheck < millis()) && IN_RUNNING_MODE) { // if not editing update work
    if (now != oldnow) { // clock is  running
      work[HR] = now.hour(); work[MIN] = now.minute();
      work[DAY] = now.day(); work[MONTH] = now.month(); work[YEAR] = now.year() - 2000;
    } else { // Use the Arduino as a backup clock
      Serial.println("RTC is not running");
      work[MIN] = work[MIN] +1;
      if (work[MIN] == 60) {
        work[MIN] = 0; work[HR] = work[HR] + 1;
      }
      if (work[HR] == 24) {
        work[HR] = 0; work[DAY] = work[DAY] + 1;
      }
      if (work[DAY] > mon[work[MONTH]]) {
        work[DAY] = 1; work[MONTH] = work[MONTH] + 1;
      }
      if (work[MONTH] > 13) {
        work[MONTH] = 1; work[YEAR] = work[YEAR] + 1;
      }
      now = DateTime(work[YEAR]+2000, work[MONTH], work[DAY], work[HR], work[MIN], 0);
      clockfail = true;
    }
    oldnow = now;
    mincheck = millis() + MINUTE;
  }  
    
 
/* ------- Aqcuire input ------- */
  
  /* read input and debounce */
  signalp = 0; signalr = 0; detect = 0; rot = 0;
  for (int i = 0; i <= BOUNCE; i++) { // eliminate bouncing
    signalp += !digitalRead(PUSH);
    signalr += !digitalRead(ROTP);
    detect += digitalRead(TWILIGHT);
    signalA = digitalRead(ROTA);
    signalB = digitalRead(ROTB);
    if (signalA > signalB) rot -= 1;
    if (signalA < signalB) rot += 1;
    delay(1);
  }
  signalp = (signalp > (BOUNCE / 2)) ? HIGH : LOW;
  signalr = (signalr > (BOUNCE / 2)) ? HIGH : LOW;
  signalt = (detect > (BOUNCE / 2)) ? true : false;
  action = 0;
  
  /* check push button */
  if (signalp == HIGH) {
    pressed = millis() + PRESS;
    while (digitalRead(PUSH) == LOW){
      ; // wait
    }
    if (millis() < pressed) {
      action = 1;  Serial.println("Action = 1");
    } else {
      action = 6;  Serial.println("Action = 6");
    }
  }

  /* check rotary push */
  if (signalr == HIGH) {
    pressed = millis() + PRESS;
    while (digitalRead(ROTP) == LOW) {
      ; // wait
    }
    if (millis() < pressed) {
      if IN_RUNNING_MODE {
        action = 2; Serial.println("Action = 2");
      } else {
        action = 5; Serial.println("Action = 5");
      }
    } else {
      action = 7; Serial.println("Action = 7");
    }
  }
  
  if (rot == 0) {
    new_signal = true;
  } else {
    if ((rot < 0) && new_signal) action = 3;
    if ((rot > 0) && new_signal) action = 4;
    if (new_signal) {
      Serial.println(" "); 
      Serial.print("Rot is:");
      Serial.println(rot); 
      Serial.print("Action = "); Serial.println(action);
    }
    new_signal = false;
  }

 /* ------- Update the display ------- */

  /* (de)activate display sleep mode */
  if (action != 0) { // If something is happening, don't go to sleep
    gotosleep = millis() + AWAKE;
    if (stat == SLEEP) { // Display is in sleep mode, wake up!
      stat = RUN; 
      lcd.display(); lcd.backlight();
      if (action != 2) action = 0;  // Manual overide must continue
    }
  }
  if (gotosleep < millis()) { // set the display to sleep
    stat = SLEEP;
    lcd.noDisplay(); lcd.noBacklight();
  }
  
  /* Refresh the display */
  if (stat == RUN) {
    lcd.setCursor(pos[0][0], pos[0][1]); lcd.print(T_SEP); lcd.print(T_FROM);
    lcd.setCursor(pos[1][0], pos[1][1]); lcd.print(T_SEP); lcd.print(T_UNTIL);
    for (i = FHR; i <= MIN; i++) {
      lcd.setCursor(pos[i][0], pos[i][1]);
      if (work[i] < 10) { //smaller than 10
        if (pos[i][2] == 59) lcd.print('0');
        if (pos[i][2] == 23) lcd.print(' ');
      }
      if (i != 7) {
        lcd.print(work[i]);
        if (pos[i][2] == 23) lcd.print(T_DP);
      }
    }
    lcd.setCursor(pos[7][0], pos[7][1]); lcd.printByte(work[MODE]);
    lcd.setCursor(0, 1); 
    
    if (work[OVER] == NORMAL) { // no manual override
    
      // write clock status
      i = ((clck) ? 4 : 5);  
      if (clockfail) i = 6;
      if ((work[MODE] == SENSOR) || (work[MODE] == OFF)) i = 1;
      lcd.printByte(i); 
      
      // write twilight status
      i = ((twilight) ? 4 : 5); 
      if(digitalRead(TWILIGHT) != twilight) i = 7; 
      if (sensorfail) i = 6;
      if ((work[MODE] == CLOCK) || (work[MODE] == OFF)) i = 1;
      lcd.printByte(i); lcd.print(T_IS);
      
      // write light status
      if IN_RUNNING_MODE {
        lcd.print((light) ? T_ON : T_OFF);
      } else {
        lcd.print(T_SET);
      }    
    } else { // Manual override
      lcd.setCursor(0, 1); 
      if (work[OVER] == OFF) {
        lcd.print(T_OFF);
      } else {
        lcd.print(T_ON);
      }    
      lcd.print("   ");
    }
  }
  
/* ------- Process all input ------- */
  
  /* action 6 Edit/adjust date */
  if (action == 6) {
    if IN_RUNNING_MODE { 
      Serial.println("Start editing date, stat=10"); 
      stat = DAY; setdate = true; // we are in the second editing screen
      lcd.setCursor(0,0); lcd.print("                ");
      lcd.setCursor(0,1); lcd.print("                ");
      lcd.setCursor(0,0); 
      if (work[DAY] < 10) lcd.print(' '); lcd.print(work[DAY]); lcd.print(T_DAT); 
      if (work[MONTH] < 10) lcd.print('0'); lcd.print(work[MONTH]); lcd.print(T_DAT); 
      lcd.print(work[YEAR]);
      lcd.print(T_COR); 
      lcd.setCursor(0,1); 
      if (work[STATUS] == NONE) lcd.print(T_SAM);
      if (work[STATUS] == WINTER) lcd.print(T_WIN);
      if (work[STATUS] == SUMMER) lcd.print(T_SUM);
      if (work[STATUS] == INTER) lcd.print(T_INT);
      lcd.setCursor(12,1); 
      if (abs(work[DRIFT] - ZERO) < 10) lcd.print(" ");
      if (work[DRIFT] == ZERO) lcd.print(" ");
      if (work[DRIFT] > ZERO) lcd.print("+");
      if (work[DRIFT] >= ZERO) {
        lcd.print((work[DRIFT] - ZERO));
        lcd.print("s");
      } else { 
        lcd.print((work[DRIFT] - ZERO));
        lcd.print("s");
      }
      lcd.setCursor(0,0); lcd.cursor_on(); 
    } else { 
      action = 1;
    }
  }

  /* action 1 - toggle settings/run */
  if (action == 1) {
    
    /* start editing settings */
    if IN_RUNNING_MODE { // Goto editing mode
      Serial.println("Start editing settings, stat=3");
      stat = 3; 
      lcd.setCursor(0,1); lcd.print(T_SET);
      lcd.setCursor(7,0); lcd.cursor_on(); lcd.print(T_FROM);
      adjusted = false;
    } else { // in EDITING MODE, save settings and go to running mode
    
    /* save changed settings */
      Serial.print("EEPROM pt ="); Serial.println(pt); 
      lcd.cursor_off();
      
    /* the date may have been changed, if so save */
      if (setdate) { // 2nd settings screem
        if ((work[DAY] + work[MONTH] + work[YEAR] + 2000) !=  (now.day() + now.month() + now.year())) { // date has changed
         Serial.print("Save changed date settings, stat="); Serial.println(stat);    
         Serial.print("Date was: "); Serial.print(now.day()); Serial.print(T_DAT); Serial.print(now.month()); Serial.print(T_DAT); Serial.println(now.year());
          RTC.adjust(DateTime(work[YEAR]+2000, work[MONTH], work[DAY], now.hour(), now.minute(), now.second())); // adjust the date
          now = RTC.now(); 
          Serial.print("Date is: "); Serial.print(now.day()); Serial.print(T_DAT); Serial.print(now.month()); Serial.print(T_DAT); Serial.println(now.year());
        }
        setdate = false;
 
    /* the time may have been changed, if so save */
      } else { // 1st settings screem
        if (adjusted) {
          check1 = -((now.hour() * 60) + now.minute());
          Serial.print("Clock was: "); Serial.print(now.hour()); Serial.print(T_DP); Serial.println(now.minute());
          RTC.adjust(DateTime(now.year(), now.month(), now.day(), work[HR], work[MIN], 0)); // if adjusted set clock to new time
          now = RTC.now(); 
          Serial.print("Clock is:  "); Serial.print(now.hour()); Serial.print(T_DP); Serial.println(now.minute());
          check2 = (now.unixtime() / 86400L) - (lastadjust.unixtime() / 86400L);  // number of days between two adjustments
          check1 += (now.hour() * 3600) + (now.minute() * 60); // the number of seconds the clock has been adjusted + is foreward, - is backward
          /* Check if clock speed need to be adjusted */   
          if ((lastadjust.year() != 2000) && (abs(check1) <= (2*ZERO)) && (check2 >= DRIDAY)) { // if clock was adjusted at least once, more than DRIDAY days ago and less than 20min: adjust clock
            Serial.print("Last adjustment: "); Serial.print(now.day()); Serial.print(T_DAT); Serial.print(now.month()); Serial.print(T_DAT); Serial.println(now.year()); 
            Serial.print("                 "); Serial.print(now.hour()); Serial.print(T_DP); Serial.println(now.minute()); 
            Serial.print("That is: "); Serial.print(check2); Serial.println(" days"); 
            Serial.print("Adjusted: "); Serial.print(check1); Serial.println(" seconds per day"); 
            work[DRIFT] += (check2 / check1);
            Serial.print("New drift correction: "); Serial.print(work[DRIFT]); Serial.println(" seconds per day"); 
          }
          lastadjust = now;
        }
      }
      Serial.print("Save changed settings, stat="); Serial.println(stat);
      work[HASH] = work[FHR] + work[FMIN] + work[UHR] + work[UMIN] + work[MODE] + work[STATUS] + work[DRIFT]; // calculate the new checksum
      for (i = 0; i < SIZE; i++) { // fill the EEPROM with work[]
        if (pos[i][STOR] != 0) { // value need to be stored in EEPROM
          EEPROM.update(pt+pos[i][STOR], work[i]); 
          if (EEPROM.read(pt+pos[i][STOR]) != work[i]) rewrite = true;
        }    
      }
      work[FROM] = (work[FHR] * 60) + work[FMIN]; 
      work[UNTIL] = (work[UHR] * 60) + work[UMIN];
      stat = RUN; 
    }
  }

  /* action 3/4  Change value */
  if (((action == 3) || (action == 4)) && IN_EDITING_MODE) {  // process rotary knob
    Serial.print("Work["); Serial.print(stat); Serial.print("] was "); Serial.println(work[stat]);
    if (action == 3) work[stat] = work[stat] - 1;
    if (action == 4) work[stat] = work[stat] + 1;      
    if ( work[stat] < 0) work[stat] = pos[stat][MAX]; // if lower than minimum set maximum
    if ( work[stat] > pos[stat][MAX]) work[stat] = 0; // if higher than maximum set minimum
    Serial.print("Work["); Serial.print(stat); Serial.print("] is "); Serial.println(work[stat]);

    if ((stat == HR) || (stat == MIN)) adjusted = true; // notice that clock has been adjusted
    lcd.setCursor(pos[stat][0], pos[stat][1]);
    switch (stat) {
    case MODE:
      lcd.printByte(work[MODE]);
      break;
    case STATUS:
      if (work[stat] == NONE) lcd.print(T_SAM);
      if (work[stat] == WINTER) lcd.print(T_WIN);
      if (work[stat] == SUMMER) lcd.print(T_SUM);
      break;
    case DRIFT:
      if (abs(work[DRIFT] - ZERO) < 10) lcd.print(" ");
      if (work[DRIFT] == ZERO) lcd.print(" ");
      if (work[DRIFT] > ZERO) lcd.print("+");
      if (work[DRIFT] >= ZERO) {
        lcd.print((work[DRIFT] - ZERO));
        lcd.print("s");
      } else { 
        lcd.print((work[DRIFT] - ZERO));
        lcd.print("s");
      }
      break;
    default:
      if (work[stat] < 10) { //smaller than 10
        if (pos[stat][MAX] == 12) lcd.print('0');
        if (pos[stat][MAX] == 23) lcd.print(' ');
        if (pos[stat][MAX] == 31) lcd.print(' ');
        if (pos[stat][MAX] == 59) lcd.print('0');
      }
      lcd.print(work[stat]);
      break;
    }
    lcd.setCursor(pos[stat][COL], pos[stat][ROW]); lcd.cursor_on();
  }

  /* action 5 Select value */
  if ((action == 5) && IN_EDITING_MODE) {
    stat += 1;
    if (setdate) {
      if (stat == FROM) stat = DAY;
    } else {
      if (stat == DAY) stat = FHR;
    }
    Serial.print("Stat="); Serial.println(stat);
    lcd.cursor_on(); lcd.setCursor(pos[stat][0], pos[stat][1]); 
  }

/* ------- Do the twilight clock thing ------- */
  
  /* check clock  */
  nu = (now.hour() * 60) + now.minute();
  clck = false;
  if ((nu >= work[FROM]) && (nu < work[UNTIL])) clck = true; // determine when the lights should be off
   
  /* check if daylight saving time needs adjustment  */
  if ((work[STATUS] != NONE) && !clockfail) { // daylight saving time must be adjusted and clock works
    i = NONE; 
    if ((REGION == "EU")) {
      temp = DateTime(now.year(), 4, 1, 1, 1, 1); // 1 april yyyy 1:01:01
      check1 = 33102L - ((temp.dayOfTheWeek() - 1) * 100); // last sunday of march march in format mmddhh
      temp = DateTime(now.year(), 11, 1, 1, 1, 1); // 1 november yyyy 1:01:01
      check2 = 103103L - ((temp.dayOfTheWeek() - 1) * 100); // last sunday of october in format mmddhh
      nu = (now.month()*100); nu = nu * 100; nu = nu + (now.day()*100) + now.hour(); // other construction gives wrong value during run-time
      i = SUMMER; if ((nu < check1) || (nu >= check2)) i = WINTER; 
    }
    // if ((REGION == "US") ) {
      // Not yet implemented, but feel free to write the code resulting in:
      // Post condition: i=1 : the clock should have winter time, i=2 : the clock should have summer time
    // }
    if ((i == WINTER) && (work[STATUS] == INTER)) { // change the clock back to winter time, final state
      work[STATUS] = WINTER;
      work[HASH] = work[FHR] + work[FMIN] + work[UHR] + work[UMIN] + work[MODE] + work[STATUS]  + work[DRIFT]; // calculate the new checksum
      EEPROM.write(pt+pos[HASH][STOR], work[HASH]); if (EEPROM.read(pt+pos[HASH][STOR]) != work[HASH]) rewrite = true; 
      EEPROM.write(pt+pos[STATUS][STOR], work[STATUS]); if (EEPROM.read(pt+pos[STATUS][STOR]) != work[STATUS]) rewrite = true; 
      Serial.print("Date="); Serial.print(now.day()); Serial.print(T_DAT); Serial.print(now.month()); Serial.print(T_DAT); Serial.println(now.year()); 
      Serial.print("Time="); Serial.print(now.hour()); Serial.print(T_DP); Serial.println(now.minute());  
      Serial.print("work[STATUS]="); Serial.println(work[STATUS]);  
    }
    if ((i == WINTER) && (work[STATUS] == SUMMER)) { // change the clock to winter time, intermediate state
      work[HR] -= 1; // this must happen once because the clock is set back
      RTC.adjust(DateTime(now.year(), now.month(), now.day(), work[HR], work[MIN], 0));
      now = RTC.now();
      work[STATUS] = INTER; // Set to three to avoid this happening more than once a year
      Serial.print("Date="); Serial.print(now.day()); Serial.print(T_DAT); Serial.print(now.month()); Serial.print(T_DAT); Serial.println(now.year()); 
      Serial.print("Winter time="); Serial.print(now.hour()); Serial.print(T_DP); Serial.println(now.minute());  
      Serial.print("work[STATUS]="); Serial.println(work[STATUS]);  
    }
    if ((i == SUMMER) && (work[STATUS] == WINTER)) { // change the clock to summer time
      work[HR] += 1; 
      RTC.adjust(DateTime(now.year(), now.month(), now.day(), work[HR], work[MIN], 0));
      now = RTC.now();
      work[STATUS] = SUMMER;
      work[HASH] = work[FHR] + work[FMIN] + work[UHR] + work[UMIN] + work[MODE] + work[STATUS]  + work[DRIFT]; // calculate the new checksum
      EEPROM.write(pt+pos[HASH][STOR], work[HASH]); if (EEPROM.read(pt+pos[HASH][STOR]) != work[HASH]) rewrite = true; 
      EEPROM.write(pt+pos[STATUS][STOR], work[STATUS]); if (EEPROM.read(pt+pos[STATUS][STOR]) != work[STATUS]) rewrite = true; 
      Serial.print("Date="); Serial.print(now.day()); Serial.print(T_DAT); Serial.print(now.month()); Serial.print(T_DAT); Serial.println(now.year()); 
      Serial.print("Summer time="); Serial.print(now.hour()); Serial.print(T_DP); Serial.println(now.minute());  
      Serial.print("work[STATUS]="); Serial.println(work[STATUS]);  
    }
  }

  /* determine if the twilight sensor works */
  if (work[MODE] == NORMAL || work[MODE] == SENSOR) { // twilight sensor is used
    if (signalt != digitalRead(TWILIGHT)) { // If the sensor changed value, then the sensor works!
      sensorfail = false;  
      failhour = now.hour() - 9; // Within 15 hours it should detect a change again
      if (failhour < 0) failhour = failhour + 24; 
    } else {
      if (now.hour() == failhour) sensorfail = true; // if it is 15 hours ago, it failed
    }
    
    /* check if twilight has occurred */
    if (!sensorfail) { // when the twilight sensor works
      if (twilight == signalt) { 
        dark = millis() + WAIT;
      } else {
        if (millis() >= dark) twilight = signalt; // if the twilight change is detected during WAIT, is is considered significant
      }  
    }      
    
    /* use the estimated twilight time as a backup for detection of twilight*/
    if (REGION == "EU") { // for the EU we can determine if the twilight is properly detected
      nu = (now.month() - 1);
      nu = (nu * 31) + now.day();
      if (nu > 186) nu = nu - 186;
      nu = int(272 * (nu / 186)); // 6 months * 31 days  is 186, difference between early and late is 272 minutes
      if (now.month() < 7) { // calculate the estimated time twilight should be detected 
        check1 = 529 - nu + MARGIN; // latest light is 8:49 winter time + 30 minutes margin
        check2 = 990 + nu + MARGIN; // earliest dark is 16:30 winter time + 30 minutes margin
      } else {
        check1 = 260 + nu + MARGIN; //  earliest light is 4:20 winter time + 30 minutes margin
        check2 = 1264 - nu + MARGIN; // latest dark is 21:04 winter time + 30 minutes margin
      }  
      nu = ((now.hour() * 60) + now.minute()); // minutes after midnight 
      if (work[STATUS] == SUMMER) nu = nu - 60; // if summer time, change to winter time
      if (sensorfail) twilight = false;
      if ((nu > check2) && (nu < check1) && (!twilight)) { // no darkness detected and it should be dark
        sensorfail = true;
        twilight = true; // switch the lights on!
      }
    }
  }  
      
  /* action 2 Manual override */
  if (action == 2) { // manual switch on/off by setting work[OVER] variable
    Serial.print("Old work[OVER]=");  Serial.print(work[OVER]);
    switch (work[OVER]) {
    case NORMAL:
      work[OVER] = ON;
      break;
    case OFF:
      work[OVER] = NORMAL;
      break;
    case ON:
      work[OVER] = OFF;
      break;
    }
    Serial.print(", new work[MODE]=");  Serial.print(work[MODE]);
    Serial.print(", work[OVER]=");  Serial.println(work[OVER]);
  }
   
  /* determine if the lights need to be switched on or off */
  if (work[MODE] == NORMAL) newlight = (twilight && clck);
  if (work[MODE] == OFF) newlight = false;
  if (work[MODE] == CLOCK) newlight = clck;
  if (work[MODE] == SENSOR) newlight = twilight;
  
  if (work[OVER] == NORMAL) light = newlight;
  if (work[OVER] == OFF) light = false;
  if (work[OVER] == ON) light = true;
   
  // determine the end of manual override
  if ((lastlight != newlight) && (work[OVER] == ON) && (newlight == true)) work[OVER] = NORMAL;
  if ((lastlight != newlight) && (work[OVER] == OFF)) work[OVER] == NORMAL;  
  lastlight = newlight;
    
  /* Switch the lights */    
  digitalWrite(LIGHTON, light ? HIGH : LOW);
     
  /* Repair a broken EEPROM */
  if (rewrite) { 
    pt += ESIZE; rewrite = false;
    work[HASH] = work[FHR] + work[FMIN] + work[UHR] + work[UMIN] + work[MODE] + work[STATUS] + work[DRIFT];
    EEPROM.update(1, pt); // location of data in EEPROM
    Serial.print("Rewrite EEPROM with pointer=");  Serial.println(pt);
    for (i = 0; i < SIZE; i++) { // fill the EEPROM with work[]
      if (pos[i][STOR] != 0) { // value need to be stored in EEPROM
        EEPROM.update(pt+pos[i][STOR], work[i]); 
        if (EEPROM.read(pt+pos[i][STOR]) != work[i]) rewrite = true;
      }    
    }
  } 
}
